// AliBaba.C

#include "AliBaba.H"
#include "Camera.H"
#include "Carpet.H"
#include "Ground.H"
#include "Cave.H"
#include "Light.H"
#include "Fog.H"
#include "TargetReticle.H"
#include "CocoBomb.H"
#include "Explosion.H"
#include "Particles.H"
#include "Rider.H"
#include "FPSView.H"
#include "ScoreView.H"
#include "PauseView.H"
#include "Radar.H"

#include <string.h>


namespace CS148 {

static const unsigned int targetFrameRate = 30;


class Animator : public Functor
{
public:
  Animator ( AliBaba* app, unsigned int fps ) 
    : app(app), msecs((unsigned int)1000.0 / fps), timer(),
      delayAccum(0), frameCounter(0), paused(false)
  {
    app->after ( msecs, this );
  }

  ~Animator () {
    app->cancelAfter ( this );
  }

  void func () {
    Real elapsed = timer.getElapsed(true);

    if ( !paused )
      app->animate( elapsed );

    app->after ( msecs, this );
//     delayAccum += msecs;
//     frameCounter++;
//     if ( frameCounter == 50 ) {
//       cerr << "anim fps = " << (frameCounter / (delayAccum * 1e-3)) << "\n";
//       frameCounter = 0;
//       delayAccum = 0;
//     }
  }

  bool isPaused() const { return paused; }
  void pause () { paused = true; }
  void resume () { paused = false; }


private:
  AliBaba* app;
  unsigned int msecs;
  Timer timer;

  unsigned int delayAccum;
  int frameCounter;

  bool paused;
};




AliBaba::AliBaba ()
  : GlutApp(),
    slowMode(false),
    fpsView(new FPSView()), anim(0),
    ground(0), cave(0), carpet(0), camera(0),
    reticle(0), score(0), radar(0), pauseView(0), 
    movers(), anims(), riders()
{

}


AliBaba::~AliBaba ()
{
  cancelAfter ( this );

  delete anim;
  delete camera;
  delete ground;
  delete cave;
  delete carpet;

  for ( vector<Rider*>::iterator iter = riders.begin();
	iter != riders.end(); ++iter )
    delete *iter;
}



// ----------------------------------------------------------------------
// Initialization


void AliBaba::processArgs ( int& argc, char** argv )
{
  for ( int i=1; i < argc; ++i ) {
    if ( !strcmp ( argv[i], "-slow" ) ) slowMode = true;

    else {
      cerr << "Unrecognized arg: " << argv[i] << "\n";
      exit(2);
    }
  }
}


void AliBaba::init ()
{
  GlutApp::init();

  srand ( (unsigned int) time(0) );

  setFrameRate ( targetFrameRate );

  glLightModeli ( GL_LIGHT_MODEL_LOCAL_VIEWER, GL_TRUE );

  glEnable ( GL_ALPHA_TEST );
  glEnable ( GL_CULL_FACE );
  glFrontFace ( GL_CCW );
  glCullFace ( GL_BACK );

  if ( slowMode ) {
    glHint ( GL_POINT_SMOOTH_HINT, GL_FASTEST );
    glHint ( GL_LINE_SMOOTH_HINT, GL_FASTEST );
    glHint ( GL_POLYGON_SMOOTH_HINT, GL_FASTEST );
    glHint ( GL_FOG_HINT, GL_FASTEST );
    glHint ( GL_PERSPECTIVE_CORRECTION_HINT, GL_FASTEST );
  } else {
    glHint ( GL_POINT_SMOOTH_HINT, GL_NICEST );
    glHint ( GL_LINE_SMOOTH_HINT, GL_NICEST );
    glHint ( GL_POLYGON_SMOOTH_HINT, GL_NICEST );
    glHint ( GL_FOG_HINT, GL_NICEST );
    glHint ( GL_PERSPECTIVE_CORRECTION_HINT, GL_NICEST );

    //    glLightModeli ( GL_LIGHT_MODEL_TWO_SIDE, GL_TRUE );
  }


  clear ( Color ( .8, .7, .6 ) ); // Same as haze color, for slow mode

  ground = new Ground(400,2000, 2.5, slowMode );
  cave = new Cave ( this, ground, slowMode );
  carpet = new Carpet( ground, slowMode );
  camera = new Camera( carpet, slowMode );

  // Set up the camera
  addViewport ( camera );
  camera->setPerspective ( 60, 0, 0.1, 100.0 );

  addViewport ( fpsView );
  fpsView->func();

  reticle = new TargetReticle ( 0, 0, getScreenWidth(), getScreenHeight() );
  addViewport ( reticle );

  score = new ScoreView ( 0, 0, getScreenWidth(), getScreenHeight() );
  addViewport ( score );
   
  radar = new Radar ( 15, 15, 175, 175, carpet, riders );
  addViewport ( radar );

  pauseView = new PauseView ( this,
			      0, 0, getScreenWidth(), getScreenHeight() );

  // Create the environment
  clear();
  buildWorld();

  // Set it all in motion
  anim = new Animator(this, targetFrameRate);
}


void AliBaba::reshape ( int w, int h )
{
  glMatrixMode(GL_MODELVIEW);
  glLoadIdentity();
  reticle->reshaped ( 0, 0, w, h );
  score->reshaped ( 0, 0, w, h );
  pauseView->reshaped ( 0, 0, w, h );

  glLoadIdentity();
  camera->reshaped ( 0, 0, w, h );
}


void AliBaba::display ()
{
  GlutApp::display();
}


// ----------------------------------------------------------------------
// Display and Animation


void AliBaba::func ()
{

}






// ----------------------------------------------------------------------
// Keyboard, mouse handling
  
void AliBaba::keyboard ( const KeyEvent& ev )
{
//   if ( ev.down ) cerr << "key = " << int(ev.c) << " = " << ev.c << "\n";

  switch ( ev.c ) {

  // Steering
  case 'a': case 'A':
  case 's': case 'S':
  case 'w': case 'W':
  case 'd': case 'D':
    if ( !anim->isPaused() ) carpet->key ( ev );
    break;

  // Warping the carpet / camera for debugging
  case '7': carpet->warp ( 50, 50 ); break;
  case '8': carpet->warp ( 50, ground->getLength()-50 ); break;
  case '9': carpet->warp ( ground->getWidth()-50,
			   ground->getLength()-50 ); break;
  case '0': carpet->warp ( ground->getWidth()-50, 50 ); break;
  case ' ': cave->openSesame(); break;

  case 'c':
    carpet->warp ( ground->getWidth()/2-20, ground->getLength()-80 );
    carpet->setVelo ( Vec3 ( 0.003, 0, .01 ) );
    break;

//   case 'q':
//   case 'Q':
//     exit ( 0 );
//     break;

//   case 'p':
//     cerr << "carpet at " << carpet->getPos() << "\n";
//     break;

  case 27:
    if ( ev.down ) {
      if ( anim->isPaused() ) {
	anim->resume();
	removeViewport ( pauseView );
      } else { 
	anim->pause();
	addViewport ( pauseView );
      }
    }
    break;

  case 0:
    switch ( ev.special ) {


    }
    break;

  default:
    break;
  }

}
 
 
// Get mouse button events
void AliBaba::button ( const ButtonEvent& ev )
{
  if ( anim->isPaused() ) {
    if ( ev.button ==1 && ev.down ) {
      pauseView->pick ( ev );
    }
  } else {
    if ( ev.button == 1 && ev.down ) {
      launchBomb ( ev.x, ev.y );
    }
  }

}


// Get mouse motion events
void AliBaba::motion ( const MotionEvent& ev )
{
  if ( anim->isPaused() ) 
    pauseView->motion ( ev );
  else
    reticle->motion ( ev );
}


// ----------------------------------------------------------------------
// Building the world


void AliBaba::buildWorld ()
{
  // Reset world
  clearDisplayLists();
  clearPosObjs();
  movers.clear();
  anims.clear();

  glMatrixMode ( GL_MODELVIEW );
  glPushMatrix();
  glLoadIdentity();


  // Let there be light
  Light::setGlobalAmbient ( Vec3( .5, .5, .5 ) );
  glPushMatrix();
//   Light sun(0, Vec4( 200, 10000, 1000, 1 ) );
  Light sun(0, Vec4( 0, 10000, 0, 1 ) );
  sun.setDiffuse ( Vec3(1, 1, .8 ) );
  sun.setAmbient ( Vec3(1, 1, .8 ) );
  sun.setSpecular ( Vec3(.6, .6, .6) );
  sun.setSpot ( 180, Vec3 ( 0, -1, 0 ) );
  //sun.setAttenuation ( .2 );
  addDisplayList ( sun.buildNew() );
  glPopMatrix();


  // Add ground
  GroundPatch* top = ground->getTopPatch();
  addPosObj ( top );
  Real boundary = slowMode? 30 : 10;

  // Fog to obscure for clipping plane, looks like sand haze
  if ( !slowMode ) {
    Fog haze ( GL_LINEAR, Color ( .8, .7, .6 ), .01, 50, 130 );
    addDisplayList ( haze.buildNew( true ) );
  }


  // Add the carpet
  carpet->buildFrames();
  carpet->setPos ( Vec3( ground->getWidth()/2, 10, 20 ) );
  carpet->setVelo ( Vec3 ( 0, 0, 5 ) );
  carpet->setBoundary ( boundary, boundary,
			ground->getWidth()-boundary,
			ground->getLength()-boundary );
  addMover ( carpet );
  addAnim ( carpet );
  camera->followCarpet();

  // The riders
  addMover ( cave );
  for ( int i=slowMode? 15 : 40; i--; ) {
    placeRider( boundary );
  }

  glPopMatrix();
}


void AliBaba::placeRider ( Real boundary )
{
  Rider* rider = new Rider ( this, ground, cave );
  rider->buildFrames();
  addMover ( rider );
  addAnim ( rider );
  riders.push_back ( rider );

  // Choose a starting position
  // Randomly chosen near the start of the ground
//   rider->setPos ( Vec3( ground->getWidth()/2, 5, 30 ) );
  Real x = 5 + (ground->getWidth()-10) * 0.01 * (rand() %100);
  Real z = 5 + 100 * 0.01 * (rand()%100);
  Real y = ground->getHeight ( x, z );
  rider->setPos ( Vec3( x, y, z ) );
  rider->setVelo ( Vec3( 0, 0, 10 + (rand()%15) ) );
  rider->setBoundary ( boundary, boundary,
		       ground->getWidth()-boundary,
		       ground->getLength()-boundary );

  score->addedRider();
}



// ----------------------------------------------------------------------
// Game events


void AliBaba::addMover ( Mover* mv )
{
  addPosObj ( mv );

  for ( vector<Mover*>::iterator iter = movers.begin();
	iter != movers.end(); ++iter ) {
    if ( (*iter) == 0 ) {
      *iter = mv;
      return;
    }
  }
  movers.push_back ( mv );

}


void AliBaba::removeMover ( Mover* mv )
{
  removePosObj ( mv );

  for ( vector<Mover*>::iterator iter = movers.begin();
	iter != movers.end(); ++iter ) {
    if ( (*iter) == mv ) {
      *iter = 0;
      //movers.erase ( iter ); <-- prevent removal while iterating
      return;
    }
  }

}


void AliBaba::addAnim ( Animated* anim )
{
  for ( vector<Animated*>::iterator iter = anims.begin();
	iter != anims.end(); ++iter ) {
    if ( (*iter) == 0 ) {
      *iter = anim;
      return;
    }
  }
  anims.push_back ( anim );
}


void AliBaba::removeAnim ( Animated* anim )
{
  for ( vector<Animated*>::iterator iter = anims.begin();
	iter != anims.end(); ++iter ) {
    if ( (*iter) == anim ) {
      *iter = 0;
      //anims.erase ( iter ); <-- prevent removal while iterating
      return;
    }
  }
  
}


void AliBaba::animate ( Real interval )
{
  for ( vector<Mover*>::iterator iter = movers.begin();
	iter != movers.end(); ++iter )
    if ( *iter ) (*iter)->move ( interval );

  for ( vector<Animated*>::iterator iter = anims.begin();
	iter != anims.end(); ++iter )
    if ( *iter ) (*iter)->advanceFrame ( interval );

  camera->followCarpet();
}


void AliBaba::launchBomb ( int x, int y )
{
  // ------------------------------------------------------------
  // Determine vector toward the target reticle

  Vec3 velo = carpet->getVelo();
  Vec3 rpos = reticle->getPos();

  // Reticle position translates to a rotation (phi, around y-axis)
  // and elevation angle (phi, around z-axis)

  // phi can go between -45 and 45 degrees
  Real pctx = rpos[0]/getScreenWidth();
  Real phi = -45.0 + 90.0*pctx;
  phi += carpet->getRotation();

  // theta can go between 45 and 120 degrees (0 is straight up)
  Real pcty = 1.0 - rpos[1]/getScreenHeight();
  Real theta = 45.0 + 75.0*pcty;

  // Bombs thrown with constant velocity
  Real speed = 20; // ~40 mph / 70kmh

  Vec3 launchsp = Vec3(speed, theta, phi );
  Vec3 launch = sphere2cart ( launchsp );
//   cerr << "launch vector = " << launch 
//        << " (" << Vec3(speed, theta, phi) << " )"
//        << " pctx=" << pctx << " pcty=" << pcty << "\n";
  velo += launch;

  // Create the bomb and set it in motion
  CocoBomb* cb = new CocoBomb ( this, ground );
  cb->setPos ( carpet->getPos() );
  cb->setVelo ( velo );
  addMover ( cb );
}



void AliBaba::exploded ( CocoBomb* cb )
{
  // Deform the ground...
  Vec3 pos = cb->getPos();
  pos[1] = ground->getHeight ( pos[0], pos[2] );
  const Vec3& velo = cb->getVelo();
  ground->explosion ( pos[0], pos[2] );

  new Explosion ( this, pos );
  
  // Check the riders
  vector<Rider*> hitRiders;
  for ( vector<Rider*>::iterator iter = riders.begin();
	iter != riders.end(); ++iter ) {
    Rider* rd = *iter;

    Vec3 dist = rd->getPos() - pos;
    Real l = len(dist);

    if ( l <= 10.0 ) {
      // Kaboom!
      hitRiders.push_back ( rd );
    }
    else if ( l <= 30.0 ) {
      rd->nearMiss();
    }
  }
  for ( vector<Rider*>::iterator iter = hitRiders.begin();
	iter != hitRiders.end(); ++iter ) {
    new Particles ( this, (*iter)->getPos(), velo );
    (*iter)->hitByBomb();
  }


  // Get rid of the bomb...
  removeMover ( cb );
  delete cb;
}



void AliBaba::killed ( Rider* rd )
{
  removeRider ( rd );
  score->killedRider();
}


void AliBaba::reachedCave ( Rider* rd )
{
  removeRider ( rd );
  score->escapedRider();
}


void AliBaba::removeRider ( Rider* rd )
{
  for ( vector<Rider*>::iterator iter = riders.begin();
	iter != riders.end(); ++iter ) {
    if ( *iter == rd ) {
      riders.erase ( iter );
      break;
    }
  }

  removeMover ( rd );
  removeAnim ( rd );
  delete rd;

}



} // namespace CS148
