// Cave.C

#include "Cave.H"

#include "AliBaba.H"
#include "Ground.H"
#include "Texture.H"
#include "Light.H"

namespace CS148 {

static const Real caveHeight = 6;

class CaveBuilder : public DisplayBuilder
{
public:
  CaveBuilder () : DisplayBuilder() {}
  ~CaveBuilder () {}

  void buildDisplayList () {
    glPushAttrib ( GL_ENABLE_BIT );
    glPushMatrix();
    glTranslatef ( 0, -3, -12 );
    glScalef ( 2, 1.5, 2 );

    // Dark inside of the cave
    static const Vec4 glow ( 0, 0, .4, 1 );
    static const Vec4 dark ( 0, 0, 0, 1 );
    glMaterialfv ( GL_FRONT_AND_BACK, GL_AMBIENT_AND_DIFFUSE, glow );
    glMaterialfv ( GL_FRONT_AND_BACK, GL_SPECULAR, glow );
    glMaterialfv ( GL_FRONT_AND_BACK, GL_EMISSION, glow );
    glMaterialf ( GL_FRONT_AND_BACK, GL_SHININESS, 120 );

    glCullFace ( GL_FRONT );
    glutSolidCone ( caveHeight-1, 6, 10, 1 );
    glCullFace ( GL_BACK );

    glMaterialfv ( GL_FRONT_AND_BACK, GL_EMISSION, dark ); // Turn off glow

    // Walls of the cave
    static const Vec4 wallAmbient ( .4, .2, .2, 1.0 );
    glMaterialfv ( GL_FRONT_AND_BACK, GL_AMBIENT_AND_DIFFUSE, wallAmbient );
    glMaterialfv ( GL_FRONT_AND_BACK, GL_SPECULAR, wallAmbient );
    glMaterialf ( GL_FRONT_AND_BACK, GL_SHININESS, 10 );

//     assert ( GroundPatch::wallTexture() );
//     GroundPatch::wallTexture()->bind();

    glTranslatef ( 0, 0, 1 );
    glScalef ( 1, 1, 5 );
    glutSolidTorus ( 2, caveHeight/2+3, 6, 16 );

//     GroundPatch::wallTexture()->unbind();

    glPopAttrib();
    glPopMatrix();
  }
};



Cave::Cave ( AliBaba* app, Ground* ground, bool slowMode )
  : Mover(), Functor(),
    app(app), ground(ground),
    state(closed), minHeight(0), maxHeight(0), targetHeight(0),
    dlist(0), l1(0), l2(0), l3(0)
{
  Real cx = ground->getWidth()/2;
  Real cz = ground->getLength()-10;
  Real cy = ground->getMinHeight(cx, cz,15) - caveHeight;
  setPos ( Vec3( cx, cy, cz ) );

  maxHeight = ground->getMinHeight ( cx, cz, 15 );
  minHeight = maxHeight - caveHeight;

  setBoundRadius ( caveHeight * 4 );

  l1 = new Light(1, Vec4 ( 0, .5, -7, 1 ) );
  l1->setDiffuse ( Vec3( 0, 0, 1 ) );
  l1->setSpecular ( Vec3( 0, 0, 1 ) );
  l1->setSpot ( 60, Vec3( 0, -1, -.6 ) );

  l2 = new Light(2, Vec4 ( 0, 7, -7, 1 ) );
  l2->setDiffuse ( Vec3( 0, 0, 1 ) );
  l2->setSpecular ( Vec3( 0, 0, 1 ) );
  l2->setSpot ( 60, Vec3( 0, -1, -.1 ) );
}


Cave::~Cave ()
{
  app->cancelAfter ( this );
  delete dlist;
}


void Cave::render ( Viewport* vp )
{
  if ( !vp->canSee( this ) ) {
    l1->disable();
    l2->disable();
    return;
  }

  if ( !dlist ) {
    dlist = CaveBuilder().buildNew(true);
  }

  glPushMatrix();
  translate();
  dlist->call();
  Real h = getPos()[1];
  if ( state != closed ) {
    // Light emitting a glow inside the cave
    l1->enable();
    l2->enable();

    l1->setAttenuation ( .15, 0, .0005 +
			 .01*(1.0 - (h-minHeight)/caveHeight) ); 
    l2->setAttenuation ( .15, 0, .001 +
			 .01*(1.0 - (h-minHeight)/caveHeight) ); 
    l1->draw();
    l2->draw();

  } else {
    l1->disable();
    l2->disable();
  }

  glPopMatrix();
}


void Cave::openSesame ()
{
  switch ( state ) {
  case closed:
  case closing: {
    state = opening;
    setVelo ( Vec3 ( 0, 2, 0 ) );
    app->after ( 10000, this );
    targetHeight = maxHeight;
  } break;

  case opened:
  case opening:
    break;
  }

}


void Cave::move ( Real interval )
{
  Mover::move ( interval );

  switch ( state ) {
  case closed:
  case opened:
    break;

  case closing:
    if ( getPos()[1] <= targetHeight ) {
      state = closed;
      setVelo ( Vec3 ( 0, 0, 0 ) );
    }
    break;

  case opening:
    if ( getPos()[1] >= targetHeight ){
      state = opened;
      setVelo ( Vec3 ( 0, 0, 0 ) );
    }
    break;
  }
}


void Cave::func ()
{
  switch ( state ) {
  case opened:
  case opening: {
    state = closing;
    setVelo ( Vec3( 0, -2, 0 ) );
    targetHeight = minHeight;
  } break;

  case closed:
  case closing:
    break;
  }

}


} // namespace CS148


