// GlutApp.C
// **********************************************************************
// Base class defining a GLUT-based application
// Author: Jeremy Slade
// **********************************************************************

#include "GlutApp.H"

#include <cstring>
#include <cstdio>
#include <cstdlib>
#include <iostream>


using namespace std;

namespace CS148 {


// The singleton instance:
static GlutApp* single = 0;



/**********************************************************************
 * GlutApp Funcs
 **********************************************************************/

class Funcs {
public:
  static void idle ( void ) {
    single->_idle();
  }
  
  static void display ( void ) {
    single->_display();
  }
  
  static void reshape ( int w, int h ) {
    single->_reshape ( w, h );
  }
  
  static void keyup ( unsigned char c, int x, int y ) {
    single->_keyboard ( c, 0, x, y, false  );
  }

  static void keydown ( unsigned char c, int x, int y ) {
    single->_keyboard ( c, 0, x, y, true  );
  }

  static void speckeyup ( int sp, int x, int y ) {
    single->_keyboard ( 0, sp, x, y, false  );
  }

  static void speckeydown ( int sp, int x, int y ) {
    single->_keyboard ( 0, sp, x, y, true  );
  }

  static void button ( int b, int s, int x, int y ) {
    single->_button ( b, s, x, y );
  }
  
  static void motion ( int x, int y ) {
    single->_motion ( x, y );
  }
  
  static void passiveMotion ( int x, int y ) {
    single->_motion ( x, y );
  }
  
};




/**********************************************************************
 * GlutApp
 **********************************************************************/


/**
   Real constructor
*/
GlutApp::GlutApp ()
  : name(), width(0), height(0),
    buf2(false), depth(false),
    downButton(None), startx(0), starty(0), lastx(0), lasty(0),
    refreshing(false),
    timers(), nextTimerSlot(0),
    targetFPS(0), frameDelay(0), frameCounter(0), frameTimer(),
    displayLists(), viewports(), posObjs()
{

}


/**
   Real destructor
*/
GlutApp::~GlutApp ()
{
  clearDisplayLists();
}


/**
   Singleton constructor
*/
GlutApp* GlutApp::create ( int argc, char** argv,
			   Factory factory )
{
  if ( single )
    throw "Only a single instance can be created";
  single = factory();
  if( !single )
    throw "Factory method failed";

  glutInit ( &argc, argv );
  single->processArgs ( argc, argv );

  return single;
}



/**
   Get the existing singleton instance
*/
GlutApp* GlutApp::getInstance ()
{
  return single;
}


/**
   Process command-line args
*/
void GlutApp::processArgs ( int& /*argc*/, char** /*argv*/ )
{

}


/**
   Initialize the instance
*/
void GlutApp::init ( const string& n, int w, int h, int displayMode )
{
  name = n;
  width = w;
  height = h;

  glutInitDisplayMode ( displayMode );
  glutInitWindowSize ( width, height );

  char buf[n.size()+1];
  strcpy ( buf, n.c_str() );
  glutCreateWindow ( buf );

  buf2 = displayMode & GLUT_DOUBLE;
  depth = displayMode & GLUT_DEPTH;
  if ( depth ) glEnable(GL_DEPTH_TEST);


  // Do app-specific initialization
  init();
}


/**
   Run the application (the event loop)
*/
void GlutApp::run ( void )
{
//   glutIdleFunc ( Funcs::idle );
  glutDisplayFunc ( Funcs::display );
  glutReshapeFunc ( Funcs::reshape );

  glutIgnoreKeyRepeat ( 1 );
  glutKeyboardFunc ( Funcs::keydown );
  glutKeyboardUpFunc ( Funcs::keyup );
  glutSpecialFunc ( Funcs::speckeydown );
  glutSpecialUpFunc ( Funcs::speckeyup );

  glutMouseFunc ( Funcs::button );
  glutMotionFunc ( Funcs::motion );
  glutPassiveMotionFunc ( Funcs::passiveMotion );

  glutMainLoop();
}


/**
   Do app-specific initialization.  
*/
void GlutApp::init ( void )
{


}


/**
   Idle function registered as glutIdleFunc()
*/
void GlutApp::_idle ( void )
{
  idle();
}


/**
   Idle function for app to implement
*/
void GlutApp::idle ( void )
{

}


/**
   Display function registered as glutDisplayFunc()
*/
void GlutApp::_display ( void )
{
// //   cerr << '.';
  display();
  flush();
  refreshing = false;
  frameCounter++;
}


/**
   Display function for app to implement
*/
void GlutApp::display ( void )
{
  clear();
  renderViewports();
}


/**
   Handle window resizing, registered as glutReshapeFunc()
*/
void GlutApp::_reshape ( int w, int h )
{
  width = w;
  height = h;

  reshape ( w, h );
  flush();
}


/**
   Reshape function for app to implement
*/
void GlutApp::reshape ( int, int )
{

}


/**
   Keyboard handler registered as glutKeyboardFunc()
*/
void GlutApp::_keyboard ( unsigned char c, int sp, int x, int y, bool down )
{
  y = fixy(y);

  keyboard ( KeyEvent(c, sp, x, y, down) );
}


/**
   Keyboard handler for app to implement
*/
void GlutApp::keyboard ( const KeyEvent& )
{


}


/**
   Mouse button handler registered as glutMouseFunc()
*/
void GlutApp::_button ( int b, int s, int x, int y )
{
  y = fixy(y);

  Button currButton = None;
  switch ( b ) {
  case GLUT_LEFT_BUTTON: currButton = Left; break;
  case GLUT_MIDDLE_BUTTON: currButton = Middle; break;
  case GLUT_RIGHT_BUTTON: currButton = Right; break;
  default: currButton = Unknown;
  }

  button ( ButtonEvent( currButton, s == GLUT_DOWN, x, y, startx, starty ) );

  // Keep track of where button events happen to
  // pass info to motion events
  if ( s == GLUT_DOWN ) {
    downButton = currButton;
  } else {
    downButton = None;
  }
  startx = lastx = x;
  starty = lasty = y;
}


/**
   Mouse button handler for app to implement
*/
void GlutApp::button ( const ButtonEvent& )
{

}


/**
   Mouse motion handler registered as 
   glutMotionFunc() and glutPassiveMotionFunc()
*/
void GlutApp::_motion ( int x, int y )
{
  y = fixy(y);

  motion ( MotionEvent ( downButton, x, y,
			 x - lastx, y - lasty, startx, starty ) );

  lastx = x;
  lasty = y;
}


/**
   Mouse motion handler for app to implement
*/
void GlutApp::motion ( const MotionEvent& )
{

}


/**
   Refresh: cause re-display to happen immediately or sometime
   in the near future
*/
void GlutApp::refresh ( bool now )
{
  if ( refreshing ) return;
  refreshing = true;
  if ( now ) _display();
  else glutPostRedisplay();
}


/**
   Flush: flush GL ops and swap buffers
*/
void GlutApp::flush ()
{
  glFlush();
  if ( buf2 ) glutSwapBuffers();
}


/**
   Clear the screen to a specific color

   Unless GL_SCISSOR_TEST is enabled and a scissor rect is set,
   will clear the entire window.  So, this should be done before
   any viewports are rendered, or it should be done individually
   per viewport after it sets the scissor rect.
*/
void GlutApp::clear ( const Color& c )
{
  glClearColor ( c.r, c.g, c.b, 0 );
  clear();
}


/**
   Clear the screen using the previous clear color
*/
void GlutApp::clear ()
{
  int mask = GL_COLOR_BUFFER_BIT;
  if ( depth ) mask |= GL_DEPTH_BUFFER_BIT;
  glClear ( mask );
}







// ----------------------------------------------------------------------
// Timers

void timerFunc ( int _slot )
{
  // Get the functor in the slot
  unsigned int slot = (unsigned int)_slot;

  Functor* f = single->timers[slot];

  // Free that slot up
  single->nextTimerSlot = slot;
  single->timers[slot] = 0;

  // Execute -- may be no functor if this
  // timer was cancelled.
  if ( f ) f->func();
}


void GlutApp::after ( unsigned int msecs, Functor* f )
{
  // Determine what slot in the timers list this goes into
  int slot = (unsigned int)nextTimerSlot;
  if ( nextTimerSlot >= timers.size() ) {
    slot = timers.size();
    timers.push_back ( f );
  } else {
    timers[nextTimerSlot] = f;
  }

  // Register the callback
  glutTimerFunc ( msecs, timerFunc, slot );


  // Next timer registered will go at end of list, unless
  // an active timer completes in the mean time and frees
  // up a slot.
  nextTimerSlot = timers.size();
}


void GlutApp::cancelAfter ( Functor* f )
{
  // Cancel the timer for the given functor.
  // Searches the timers list for all occurences of the
  // given functor, and clears that slot.
  // In the rare case that a given functor is in multiple
  // slots when it is cancelled, just means there will be
  // holes in the timers list, but that is just a small
  // ammount of wasted space.

  for ( unsigned int i=0, n=timers.size(); i < n; ++i ) {
    if ( timers[i] == f ) {
      timers[i] = 0;
      nextTimerSlot = i;
    }
  }
	
}


// ----------------------------------------------------------------------
// Target frame rate


void targetFpsFunc ( int frameDelay )
{
  glutPostRedisplay();
  glutTimerFunc ( frameDelay, targetFpsFunc, frameDelay );
}


void GlutApp::setFrameRate ( GLuint fps )
{
  targetFPS = fps;
  frameDelay = fps? 1000 / fps : 0;

  frameTimer.reset();
  frameCounter = 0;

  if ( targetFPS )
    glutTimerFunc ( frameDelay, targetFpsFunc, frameDelay );
}


GLfloat GlutApp::measureFrameRate ()
{
  GLfloat fps = frameCounter / frameTimer.getElapsed(true);
  frameCounter = 0;
  return fps;
}



// ----------------------------------------------------------------------
// Display lists


void GlutApp::addDisplayList ( DisplayList* dl )
{
  assert ( dl );
  removeDisplayList ( dl ); // Just in case
  displayLists.push_back ( dl );
}


void GlutApp::removeDisplayList ( DisplayList* dl )
{
  assert ( dl );
  for ( vector<DisplayList*>::iterator iter = displayLists.begin();
	iter != displayLists.end(); ++iter ) {
    if ( *iter == dl ) {
      displayLists.erase ( iter );
      return;
    }
  }
}


void GlutApp::clearDisplayLists ()
{
  vector<DisplayList*> oldLists = displayLists;
  displayLists.clear();
  
  for ( vector<DisplayList*>::iterator iter = oldLists.begin();
	iter != oldLists.end(); ++iter ) {
    delete *iter;
  }
}


void GlutApp::callDisplayLists ()
{
  for ( vector<DisplayList*>::iterator iter = displayLists.begin();
	iter != displayLists.end(); ++iter ) {
    DisplayList* dl = *iter;
    if ( dl ) dl->call();
  }
}



// ----------------------------------------------------------------------
// Viewport and PosObj


void GlutApp::addViewport ( Viewport* vp )
{
  viewports.push_back ( vp );
}


void GlutApp::removeViewport ( Viewport* vp )
{
  for ( vector<Viewport*>::iterator iter = viewports.begin();
	iter != viewports.end(); ++iter ) {
    if ( *iter == vp ) {
      viewports.erase ( iter );
      return;
    }
  }
}


void GlutApp::renderViewports ()
{
  for ( vector<Viewport*>::iterator iter = viewports.begin();
	iter != viewports.end(); ++iter ) {
    Viewport* vp = *iter;
    vp->makeActive();
    vp->render();
  }
}


void GlutApp::addPosObj ( PosObj* po )
{
  posObjs.push_back ( po );
}


void GlutApp::removePosObj ( PosObj* pobj )
{
  for ( vector<PosObj*>::iterator iter = posObjs.begin();
	iter != posObjs.end(); ++iter ) {
    if ( *iter == pobj ) {
      posObjs.erase ( iter );
      return;
    }
  }
}


void GlutApp::clearPosObjs ()
{
  posObjs.clear();
}


void GlutApp::renderPosObjs ( Viewport* vp )
{
  for ( vector<PosObj*>::iterator iter = posObjs.begin();
	iter != posObjs.end(); ++iter ) {
    PosObj* pobj = *iter;
    pobj->render ( vp );
  }

}



/***************************************************************************
 * DisplayList
 ***************************************************************************/


DisplayList::DisplayList ()
  : id(glGenLists(1))
{
}


DisplayList::~DisplayList ()
{
  GlutApp::getInstance()->removeDisplayList ( this );
  erase();
}


void DisplayList::call ()
{
  glCallList ( id );
}


void DisplayList::erase ()
{
  glDeleteLists ( id, 1 );
}


/***************************************************************************
 * DisplayBuilder
 ***************************************************************************/


DisplayBuilder::DisplayBuilder ()
{}


DisplayBuilder::~DisplayBuilder ()
{}


DisplayList* DisplayBuilder::buildNew ( bool identity )
{
  DisplayList* dl = newDisplayList();

  if ( identity ) {
    glPushMatrix();
    glLoadIdentity();
  }

  build ( dl, identity );

  if ( identity ) {
    glPopMatrix();
  }

  return dl;
}


DisplayList* DisplayBuilder::newDisplayList ()
{
  return new DisplayList();
}


void DisplayBuilder::build ( DisplayList* dl, bool identity )
{
  // Get the current MODELVIEW matrix,
  // So it can be loaded as part of the display list
  GLfloat mvMatrix[16];
  glGetFloatv ( GL_MODELVIEW_MATRIX, mvMatrix );

  glMatrixMode ( GL_MODELVIEW );
  glNewList ( dl->getId(), GL_COMPILE );

  glPushMatrix ();
  if ( !identity ) glMultMatrixf ( mvMatrix );
  buildDisplayList();
  glPopMatrix ();

  glEndList ();
}

/**********************************************************************
 * PosObj
 *
 **********************************************************************/

PosObj::~PosObj ()
{
  GlutApp::getInstance()->removePosObj ( this );
}


void PosObj::render ( Viewport* /*vp*/ ) {}



/**********************************************************************
 * Viewport
 *
 * Encapsulation of viewport and projection functionality,
 * allowing multiple simultaneous viewports
 *
 * Implements view-frustum culling, based on description and code
 * from: http://www.markmorley.com/opengl/frustumculling.html
 **********************************************************************/

Viewport* Viewport::activeViewport = 0;


Viewport::Viewport ( int x, int y, int w, int h )
  : PosObj(), 
    x(x), y(y), w(w), h(h),
    perspective(true), near(0), far(0),
    refObj(0), vup(Vec3(0,1,0)), saveProj(),
    frustum()
{

}


Viewport::~Viewport ()
{
  GlutApp::getInstance()->removeViewport ( this );
}


// ----------------------------------------------------------------------
// Projection

void Viewport::setArea ( int x, int y, int w, int h )
{
  this->x = x;
  this->y = y;
  this->w = w;
  this->h = h;

  if ( activeViewport == this ) activate();
}


void Viewport::getArea ( int& x, int& y, int& w, int& h ) const
{
  x = this->x;
  y = this->y;
  w = this->w;
  h = this->h;
}


void Viewport::reshaped ( int x, int y, int w, int h )
{
  setArea ( x, y, w, h );
  moved();
}


void Viewport::setPerspective ( GLdouble fov, GLdouble aspect,
				GLdouble near, GLdouble far )
{
  perspective = true;
  persp.fov = fov;
  persp.aspect = aspect;
  this->near = near;
  this->far = far;
}


void Viewport::setOrtho ( GLdouble l, GLdouble r,
			  GLdouble b, GLdouble t,
			  GLdouble near, GLdouble far )
{
  perspective = false;
  ortho.left = l;
  ortho.right = r;
  ortho.bottom = b;
  ortho.top = t;
  this->near = near;
  this->far = far;
}


void Viewport::setProjection ()
{
  // Assumes glMatrixMode(GL_PROJECTION) done previously,
  // as well as necessary push matrix, etc

  if ( perspective ) {
    GLfloat ar = persp.aspect;
    if ( ar == 0.0 ) ar = (float)w/h;
    gluPerspective ( persp.fov, ar, near, far );
//     cerr << "Did perspective fov=" << persp.fov
// 	 << " ar=" << ar
// 	 << " near=" << near
// 	 << " far=" << far << "\n";
  } else {
    glOrtho ( ortho.left, ortho.right, ortho.bottom, ortho.top,
	      near, far );
//     cerr << "Did ortho l=" << ortho.left
// 	 << " r=" << ortho.right
// 	 << " b=" << ortho.bottom
// 	 << " t=" << ortho.top
// 	 << " n=" << near
// 	 << " f=" << far
// 	 << "\n";
  }

}


void Viewport::setReference ( PosObj* p )
{
  refObj = p;
  moved();
}


void Viewport::setUp ( const Vec3& u )
{
  vup = u;
  moved();
}


void Viewport::moved ()
{
  // Called when viewport pos or ref obj pos has changed.

  static const Vec3 origin(0,0,0);

  const Vec3& eye = getPos();
  const Vec3& ref = refObj? refObj->getPos() : origin;

//   cerr << "LookAt: vp=" << this
//        << " eye=" << eye
//        << ", ref=" << ref
//        << ", vup=" << vup << "\n";

  gluLookAt ( eye[0], eye[1], eye[2],
	      ref[0], ref[1], ref[2],
	      vup[0], vup[1], vup[2] );

  extractFrustum();
    
}



// ----------------------------------------------------------------------
// Rendering


void Viewport::makeActive ()
{
  if ( activeViewport != this ) activate();
}


void Viewport::activate ()
{
//   cerr << "Active vp=" << this << " (" << x << "," << y << ") "
//         << w << "x" << h << "\n"; 
  glViewport ( x, y, w, h );
  // may want to do this at some point: glScissor ( x, y, w, h );

  // Load the projection matrix saved previously
  glMatrixMode ( GL_PROJECTION );
  glLoadIdentity();
  setProjection();
  glMatrixMode ( GL_MODELVIEW );
  glLoadIdentity();
  moved();

  activeViewport = this;
}


void Viewport::render ()
{
  GlutApp* app = single;//GlutApp::getInstance();

  // Draw the normal display lists (generally static objects)
  app->callDisplayLists();

  // Draw all the visible PosObjs (with culling)
  app->renderPosObjs( this );
}


void Viewport::renderBitmapString ( GLfloat x, GLfloat y,
				   const char* str, void* font, int spacing )
{
  for ( const char* c = str; *c; ++c ) {
    glRasterPos2f ( x, y );
    glutBitmapCharacter ( font, *c );
    x += glutBitmapWidth ( font, *c ) + spacing;
  }
}



// ----------------------------------------------------------------------
// Culling

// static const char* planeNames[6] = { "right", "left", "bottom",
// 				    "top", "far", "near" };

static void normalizePlane ( Vec4& p )
{
  p /= len ( Vec3(p[0],p[1],p[2]) );
}


/*
  Extract the six planes of the view frustum, based on the current
  modelview and projection matrices.  This should be done at each
  frame in preparation for deciding what objects to render.
*/
void Viewport::extractFrustum ()
{
  Real proj[16], modl[16], clip[16];

  glGetFloatv( GL_PROJECTION_MATRIX, proj );
  glGetFloatv( GL_MODELVIEW_MATRIX, modl );
  
  /* Combine the two matrices (multiply projection by modelview) */
  clip[0] = modl[0] * proj[0] + modl[1] * proj[4] + modl[2] * proj[8] + modl[3] * proj[12];
  clip[1] = modl[0] * proj[1] + modl[1] * proj[5] + modl[2] * proj[9] + modl[3] * proj[13];
  clip[2] = modl[0] * proj[2] + modl[1] * proj[6] + modl[2] * proj[10] + modl[3] * proj[14];
  clip[3] = modl[0] * proj[3] + modl[1] * proj[7] + modl[2] * proj[11] + modl[3] * proj[15];
  
  clip[4] = modl[4] * proj[0] + modl[5] * proj[4] + modl[6] * proj[8] + modl[7] * proj[12];
  clip[5] = modl[4] * proj[1] + modl[5] * proj[5] + modl[6] * proj[9] + modl[7] * proj[13];
  clip[6] = modl[4] * proj[2] + modl[5] * proj[6] + modl[6] * proj[10] + modl[7] * proj[14];
  clip[7] = modl[4] * proj[3] + modl[5] * proj[7] + modl[6] * proj[11] + modl[7] * proj[15];
  
  clip[8] = modl[8] * proj[0] + modl[9] * proj[4] + modl[10] * proj[8] + modl[11] * proj[12];
  clip[9] = modl[8] * proj[1] + modl[9] * proj[5] + modl[10] * proj[9] + modl[11] * proj[13];
  clip[10] = modl[8] * proj[2] + modl[9] * proj[6] + modl[10] * proj[10] + modl[11] * proj[14];
  clip[11] = modl[8] * proj[3] + modl[9] * proj[7] + modl[10] * proj[11] + modl[11] * proj[15];
  
  clip[12] = modl[12] * proj[0] + modl[13] * proj[4] + modl[14] * proj[8] + modl[15] * proj[12];
  clip[13] = modl[12] * proj[1] + modl[13] * proj[5] + modl[14] * proj[9] + modl[15] * proj[13];
  clip[14] = modl[12] * proj[2] + modl[13] * proj[6] + modl[14] * proj[10] + modl[15] * proj[14];
  clip[15] = modl[12] * proj[3] + modl[13] * proj[7] + modl[14] * proj[11] + modl[15] * proj[15];
  
  /* Extract the numbers for the RIGHT plane */
  frustum[0][0] = clip[3] - clip[0];
  frustum[0][1] = clip[7] - clip[4];
  frustum[0][2] = clip[11] - clip[8];
  frustum[0][3] = clip[15] - clip[12];
  normalizePlane( frustum[0] );
  
  /* Extract the numbers for the LEFT plane */
  frustum[1][0] = clip[3] + clip[0];
  frustum[1][1] = clip[7] + clip[4];
  frustum[1][2] = clip[11] + clip[8];
  frustum[1][3] = clip[15] + clip[12];
  normalizePlane( frustum[1] );
  
  /* Extract the BOTTOM plane */
  frustum[2][0] = clip[3] + clip[1];
  frustum[2][1] = clip[7] + clip[5];
  frustum[2][2] = clip[11] + clip[9];
  frustum[2][3] = clip[15] + clip[13];
  normalizePlane( frustum[2] );
  
  /* Extract the TOP plane */
  frustum[3][0] = clip[3] - clip[1];
  frustum[3][1] = clip[7] - clip[5];
  frustum[3][2] = clip[11] - clip[9];
  frustum[3][3] = clip[15] - clip[13];
  normalizePlane( frustum[3] );
  
  /* Extract the FAR plane */
  frustum[4][0] = clip[3] - clip[2];
  frustum[4][1] = clip[7] - clip[6];
  frustum[4][2] = clip[11] - clip[10];
  frustum[4][3] = clip[15] - clip[14];
  normalizePlane( frustum[4] );
  
  /* Extract the NEAR plane */
  frustum[5][0] = clip[3] + clip[2];
  frustum[5][1] = clip[7] + clip[6];
  frustum[5][2] = clip[11] + clip[10];
  frustum[5][3] = clip[15] + clip[14];
  normalizePlane( frustum[5] );

  
//   cerr << "Extracted frustum:\n";
//   for ( int p=0; p < 6; ++p ) {
//     cerr << planeNames[p] << ": " << frustum[p] << "\n";
//   }

}


/*
  Determine if a given object is visible - its bounding sphere
  is completely or partially within the viewing frustum.
*/
bool Viewport::canSee ( const PosObj* pobj ) const
{
  const Vec3& pos = pobj->getPos();
  GLfloat radius = pobj->getBoundRadius();
  
  GLfloat d = 0;
  for ( int p=0; p < 6; ++p ) {
    d = ( frustum[p][0] * pos[0] +
	  frustum[p][1] * pos[1] +
	  frustum[p][2] * pos[2] +
	  frustum[p][3] );
    if ( d <= -radius ) {
      // Outside this plane, so it is outside the frustum
//       cerr << "canSee(): " << pobj << " at " << pos << " behind " 
// 	   << planeNames[p] << " plane ( d=" << d 
// 	   << " p=" << frustum[p] << " )\n";
      return false;
    }
  }
  
  return true;
}



} // namespace CS148
