// GlutApp.H
// **********************************************************************
// Base class defining a GLUT-based application
// Author: Jeremy Slade
// **********************************************************************

#ifndef _GlutApp_H_
#define _GlutApp_H_

#include <GL/gl.h>
#include <GL/glu.h>
#include <GL/glut.h>

#include <string>
#include <vector>
#include <list>

#include <cstdio>
#include <cstdlib>
#include <cassert>
#include <sys/time.h>

#define VL_FLOAT
#include <svl/SVL.h>
#include <svl/SVLgl.h>


namespace CS148 {

using namespace std;


// ----------------------------------------------------------------------
// Useful functions

static const float pi_over_180 = 3.1416 / 180.0;

static inline float deg2rad ( float deg )
{
  return deg * pi_over_180;
}

static inline float rad2deg ( float rad )
{
  return rad / pi_over_180;
}

static inline float arctan ( float x, float y )
{
  if ( x == 0 )
    return ( (y == 0)? 0 : 
	     (y > 0)?  1.5708 : -1.5708 );

  float theta = atan ( fabs(y / x) );
  
  // Which quadrant is it it?
  if ( x > 0 ) {
    if ( y < 0 ) theta = 6.2832 - theta; // quad IV
    else ; // quad I
  } else /* x <= 0 */ {
    if ( y > 0 ) theta = 3.1316 - theta; // quad II
    else if ( y < 0 ) theta += 3.1416; // quad III
  }

  return theta;
}


// Conversion between coordinate systems
// Normal coord system has z up, but OpenGL convention is y up

static inline Vec3 cart2sphere ( const Vec3& v )
{
  // (x,y,z) -> (r,theta,phi)
  const Real& x = v[0];
  const Real& y = v[1];
  const Real& z = v[2];
  Real r = len(v);
  return Vec3 ( r,
		rad2deg( acos(y/r) ),
		x? rad2deg( arctan(x, z) ) : 0 );
}
  
static inline Vec3 sphere2cart ( const Vec3& v )
{
  // (r,theta,phi) -> (x,y,z)
  const Real& r = v[0];
  const Real& theta = v[1];
  const Real& phi = v[2]; 
  Real rad_theta = deg2rad(theta);
  Real rad_phi = deg2rad(phi);
  Real r_sin_theta = r * sin(rad_theta);
  Real r_cos_theta = r * cos(rad_theta);
  return Vec3 ( r_sin_theta * cos(rad_phi),
		r_cos_theta,
		r_sin_theta * sin(rad_phi) );
}


// ----------------------------------------------------------------------
// Simple color class

class Color {
public:
  Color () : r(0), g(0), b(0), a(1.0) {}
  Color ( GLfloat r, GLfloat g, GLfloat b, GLfloat a = 1.0 )
    : r(r), g(g), b(b), a(a) {}
  
  GLfloat r, g, b, a;

  operator GLfloat* () { return &r; }

  Color operator * ( GLfloat s ) { 
    return Color( r*s, g*s, b*s, a );
  }

  static Color BLACK () { return Color ( 0, 0, 0 ); }
  static Color WHITE () { return Color ( 1, 1, 1 ); }
  static Color RED () { return Color ( 1, 0, 0 ); }
  static Color GREEN () { return Color ( 0, 1, 0 ); }
  static Color BLUE () { return Color ( 0, 0, 1 ); }
};


// ----------------------------------------------------------------------
// Simple Timer class

class Timer
{
public:
  Timer () { reset(); }

  void reset() { gettimeofday ( &startTime, 0 ); }
  float getElapsed ( bool reset = false ) {
    struct timeval nowTime;
    gettimeofday ( &nowTime, 0 );
    float elapsed = ( 1.0 * (nowTime.tv_sec - startTime.tv_sec) +
		      1e-6 * (nowTime.tv_usec - startTime.tv_usec) );
    if ( reset ) startTime = nowTime;
    return elapsed;
  }

private:

  struct timeval startTime;
};



// ----------------------------------------------------------------------
// Application class 

enum Button { None=0, Left, Middle, Right, Unknown };


class Functor;

class KeyEvent;
class ButtonEvent;
class MotionEvent;

class DisplayBuilder;
class DisplayList;

class Timer;

class PosObj;
class Viewport;


class GlutApp
{
protected:
  // Singleton class, thus protected ctor/dtor
  GlutApp ();
  virtual ~GlutApp ();
  
public:

  // Singleton ctor and factory method for subclasses
  typedef GlutApp* ( Factory ) ( void );
  static GlutApp* create ( int argc, char** argv, Factory factory );
  static GlutApp* getInstance ();

  void init ( const std::string&, int width, int height,
	      int displayMode = GLUT_SINGLE | GLUT_RGBA );
  void run ( void );

  bool isDoubleBuffered () const { return buf2; }


  // App run-time attributes
  std::string getName () const { return name; }
  int getScreenWidth () const { return width; }
  int getScreenHeight () const { return height; }

  // Useful operations
  void refresh ( bool now = true );
  void flush ();

  void clear ( const Color& c );
  void clear ();

  void after ( unsigned int msecs, Functor* f ); // Timed callbacks
  void cancelAfter ( Functor* f );

  void setFrameRate ( GLuint fps );
  GLuint getFrameRate () const { return targetFPS; }
  GLfloat measureFrameRate ();


  // Display lists
  void addDisplayList ( DisplayList* dl );
  void removeDisplayList ( DisplayList* dl );
  void clearDisplayLists ();
  void callDisplayLists ();


  // Viewports / visible objects
  void addViewport ( Viewport* vp );
  void removeViewport ( Viewport* vp );
  void renderViewports ();

  void addPosObj ( PosObj* po );
  void removePosObj ( PosObj* po );
  void clearPosObjs ();
  void renderPosObjs ( Viewport* vp );

protected:
  
  // Initialization methods called as part of init(),
  // not intended to be called directly
  virtual void processArgs ( int& argc, char** argv );
  virtual void init ( void );

  // Event methods called as part of the run loop,
  // not intended to be called directly
  virtual void idle ( void );
  virtual void display ( void );
  virtual void reshape ( int w, int h );
  virtual void keyboard ( const KeyEvent& );
  virtual void button ( const ButtonEvent& );
  virtual void motion ( const MotionEvent& );


private:

  int fixy ( int y ) const { return ( height - y - 1 ); }

  void _idle ( void );
  void _display ( void );
  void _reshape ( int w, int h );
  void _keyboard ( unsigned char key, int sp, int x, int y, bool down );
  void _button ( int button, int state, int x, int y );
  void _motion ( int x, int y );

  std::string name;
  int width;
  int height;

  bool buf2; // Double buffered?
  bool depth; // Depth test enabled?

  // State for motion events
  Button downButton;
  int startx, starty, lastx, lasty;

  bool refreshing;

  // Timed callback info
  std::vector<Functor*> timers;
  unsigned int nextTimerSlot;
  friend void timerFunc ( int _slot );

  // Target frame rate
  GLuint targetFPS;
  GLuint frameDelay;
  GLuint frameCounter;
  Timer frameTimer;

  // Display lists
  std::vector<DisplayList*> displayLists;

  // Viewport and PosObj
  std::vector<Viewport*> viewports;
  std::vector<PosObj*> posObjs;

  friend class Funcs;
};



// Generic callback function object
class Functor
{
public:
  Functor () {};
  virtual ~Functor () {}

  virtual void func () = 0;
};



// ----------------------------------------------------------------------
// Event classes

class KeyEvent {
public:
  KeyEvent ( unsigned char c, int sp, int x, int y, bool down )
    : c(c), special(sp), x(x), y(y), down(down) {}

  unsigned char c;
  int special;
  int x, y;
  bool down;
};



class ButtonEvent {
public:

  ButtonEvent ( Button b, bool down, int x, int y, int sx, int sy )
    : button(b), down(down), x(x), y(y), startx(sx), starty(sy)
  {}

  Button button;
  bool down;
  int x, y;
  int startx, starty;
};


class MotionEvent {
public:
  MotionEvent ( Button b, int x, int y,
		int dx, int dy, int sx, int sy )
    : button(b), x(x), y(y),
      dx(dx), dy(dy), startx(sx), starty(sy)
  {}

  Button button;
  int x, y;
  int dx, dy;
  int startx, starty;
};



// ----------------------------------------------------------------------
// Display lists


class DisplayList 
{
public:
  DisplayList ();
  virtual ~DisplayList ();

  GLuint getId () const { return id; }
  virtual void call ();
  void erase ();

private:

  GLuint id;
};


class DisplayBuilder
{
public:
  DisplayBuilder ();
  virtual ~DisplayBuilder ();

  DisplayList* buildNew ( bool identity = false );
  void build ( DisplayList* dl, bool identity = false );
  void immediate () { buildDisplayList(); }

protected:
 
  virtual DisplayList* newDisplayList ();
  virtual void buildDisplayList () = 0;
};


// ----------------------------------------------------------------------
// Viewport


class PosObj
{
public:
  PosObj () : pos(), radius(0) {}
  virtual ~PosObj ();

  void setPos ( const Vec3& v ) { pos = v; }
  void setSpherePos ( const Vec3& v ) { pos = sphere2cart ( v ); }

  const Vec3& getPos () const { return pos; }
  Vec3 getSpherePos () const { return cart2sphere(pos); }

  void setBoundRadius ( GLfloat r ) { radius = r; }
  GLfloat getBoundRadius () const { return radius; }

  virtual void render ( Viewport* vp );
  void translate() const { glTranslatef ( pos[0], pos[1], pos[2] ); }

private:

  Vec3 pos;
  GLfloat radius;

};



class Viewport : public PosObj
{
public:
  Viewport ( int x, int y, int w, int h );
  virtual ~Viewport ();

  void setArea ( int x, int y, int w, int h );
  void getArea ( int& x, int& y, int& w, int& h ) const;
  virtual void reshaped ( int x, int y, int w, int h );


  // Setting the projection
  void setPerspective ( GLdouble fov, GLdouble aspect,
			GLdouble near, GLdouble far );
  void setOrtho ( GLdouble l, GLdouble r,
		  GLdouble b, GLdouble t,
		  GLdouble near = -1, GLdouble far = 1 );

  bool isPerspective() const { return perspective; }

  void setProjection ();

  
  // Setting the reference point
  void setReference ( PosObj* p );
  PosObj* getReference () const { return refObj; }
  void setUp ( const Vec3& u );
  void moved ();


  // Rendering
  void makeActive ();
  virtual void render ();


  static void renderBitmapString ( GLfloat x, GLfloat y,
				   const char* str, 
				   void* font = GLUT_BITMAP_HELVETICA_10,
				   int spacing = 3 );


  // Culling
  void extractFrustum ();
  bool canSee ( const PosObj* ) const;

private:

  void activate ();

  // Projection
  int x, y, w, h;
  bool perspective;
  struct Persp { GLdouble fov, aspect; };
  struct Ortho { GLdouble left, right, bottom, top; };
  union {
    Persp persp;
    Ortho ortho;
  };
  GLdouble near, far;

  PosObj* refObj;
  Vec3 vup;

  Real saveProj[16]; // Save projection matrix

  // Culling
  Vec4 frustum [6]; // Six planes bounding the frustum;


  static Viewport* activeViewport;
};



} // namespace CS148


#endif /* _GlutApp_H_ */


