// Ground.C

#include "Ground.H"
#include "Fog.H"
#include "Texture.H"

#include <iostream>
#include <iomanip>

namespace CS148 {

using namespace std;


/**********************************************************************
 * Ground
 **********************************************************************/


Ground::Ground ( GLfloat w, GLfloat h, GLfloat step, bool slowMode )
  : slowMode(slowMode), w(w), h(h),
    rows(0), cols(0), dx(0), dz(0),
    map(0), maxy(0),
    topPatch(0), expMap(0)
{
  if ( slowMode ) step *= 4;
  rows = int(h / step);
  cols = int(w / step);
  dx = step;
  dz = step;
  cerr << "rows=" << rows << " cols=" << cols
       << " dz=" << dz << " dx=" << dx << "\n";

  map = new GLfloat[rows*cols];
  expMap = new bool[rows*cols];
  makeTerrain();

}


Ground::~Ground ()
{
  delete[] map;
  delete[] expMap;
  delete topPatch;
}


GroundPatch* Ground::getTopPatch ()
{
  if ( !topPatch ) {
    topPatch = new GroundPatch ( this, 0, 0, rows-1, cols-1 );
  }
  return topPatch;
}


void Ground::rebuildPatch ( int r1, int c1, int r2, int c2 )
{
  topPatch->redraw ( r1, c1, r2, c2 );
}


void Ground::explosion ( GLfloat x, GLfloat z )
{
  int r0, c0;
  mapPos ( x, z, r0, c0 );

  // Don't deform the ground, just set explosion flag
  int radius = 1;
  for ( int r = -radius; r <= radius; ++r ) {
    for ( int c = -radius; c <= radius; ++c ) {
      int i = mapIndex( r0 + r, c0 + c );
      expMap[i] = true;
    }
  }

  radius <<= 1;
  rebuildPatch ( r0 - radius, c0 - radius,
		 r0 + radius, c0 + radius );
}


void Ground::mapPos ( GLfloat x, GLfloat z, int& r, int& c )
{
  // Closest map point:
  r = int( (z+.5) / dz );
  c = int( (x+.5) / dx );
}


void Ground::unmapPos ( int r, int c, GLfloat& x, GLfloat& z )
{
  z = r*dz;
  x = c*dx;
}


GLfloat Ground::mapAt ( GLfloat x, GLfloat z )
{
  int r, c;
  mapPos ( x, z, r, c );
  return map[mapIndex(r,c)];
}


int Ground::mapIndex ( int r, int c )
{
  while ( r < 0 ) r += rows;
  while ( c < 0 ) c += cols;
  int i = (r%rows)*cols + (c%cols);
  return i;
}


GLfloat Ground::getHeight ( GLfloat x, GLfloat z )
{
  // Height at that point
  GLfloat h = mapAt( x, z );
  GLfloat h2 = mapAt( x+dx, z+dz );
  GLfloat h3 = mapAt( x+dx, z-dz );
  GLfloat h4 = mapAt( x-dx, z+dz );
  GLfloat h5 = mapAt( x-dx, z-dz );
  return (h + h2 + h3 + h4 + h5) / 5.0;
}


GLfloat Ground::getAvgHeight ( GLfloat mx, GLfloat mz, GLfloat radius )
{
  GLfloat accum = 0.0;
  for ( GLfloat x = mx - radius; x <= mx + radius; x += radius ) {
    for ( GLfloat z = mz - radius; z <= mz + radius; z += radius ) {
      accum += getHeight ( x, z );
    }
  }

  return accum / 9;
}


GLfloat Ground::getMinHeight ( GLfloat mx, GLfloat mz, GLfloat radius )
{
  GLfloat minh = 1e6;
  for ( GLfloat x = mx - radius; x <= mx + radius; x += radius ) {
    for ( GLfloat z = mz - radius; z <= mz + radius; z += radius ) {
      GLfloat h = getHeight ( x, z );
      if ( h < minh ) minh = h;
    }
  }

  return minh;
}


void Ground::set ( int r, int c, GLfloat y )
{
  map[mapIndex(r,c)] = y;
  if ( r == 0 ) map[mapIndex(rows-1,c)] = y;
  else if ( r == rows-1 ) map[mapIndex(0,c)] = y;
  if ( c == 0 ) map[mapIndex(r,cols-1)] = y;
  else if ( c == cols-1 ) map[mapIndex(r,0)] = y;
}


void Ground::incr ( int r, int c, GLfloat dy )
{
  map[mapIndex(r,c)] += dy;
  if ( r == 0 ) map[mapIndex(rows-1,c)] += dy;
  else if ( r == rows-1 ) map[mapIndex(0,c)] += dy;
  if ( c == 0 ) map[mapIndex(r,cols-1)] += dy;
  else if ( c == cols-1 ) map[mapIndex(r,0)] += dy;
}


void Ground::makeTerrain ()
{
  cerr << "Generating terrain: " 
       << rows << " x " << cols << " ...\n";

  // Initialize the map
  for ( int i=0; i < rows*cols; ++i ) {
    map[i] = 1.0;
    expMap[i] = false;
  }

  buildDunes();
  buildWalls();

  maxy = 0;
  GLfloat miny = 100;
  for ( int r=0; r < rows; ++r )
    for ( int c=0; c < cols; ++c ) {
      if ( isWall(r,c) ) continue;
      GLfloat m = map[mapIndex(r,c)];
      if ( m > maxy ) maxy = m;
      if ( m < miny ) miny = m;
    }

  // Translate down to earth
  for ( int i=0; i < rows*cols; ++i ) map[i] -= miny;
  maxy -= miny;

  cerr << "Terrain generated\n";
}


static const Real wallHeight = 15;
static const int wallWidth = 6;
bool Ground::isWall ( int r, int c ) const
{
  return ( (c%cols) < wallWidth || (c%cols) > (cols+1-wallWidth) ||
	   (r%rows) < wallWidth || (r%rows) > (rows+1-wallWidth) );  
}



void Ground::buildWalls ()
{
  // Want the edges of the map to be steep walls:

  // The walls slope in steeply
  Real wh = wallHeight, dh = 3.0;
  for ( int i=0; wh > 0; ++i, wh -= dh ) {
    for ( int r=0; r < rows; ++r ) {
      incr ( r, i, wh );
      incr ( r, cols-1-i, wh );
    }
    for ( int c=0; c < cols; ++c ) {
      incr ( i, c, wh );
      incr ( rows-1-i, c, wh );
    }
  }

}


void randomRowCol ( int rows, int cols, int& r, int& c )
{
  static const int margin = 0;//10;
  
  r = margin + (rand() % (rows - 2*margin));
  c = margin + (rand() % (cols - 2*margin));
}


void Ground::buildDunes ()
{
  // Add rolling dunes:
  // for N times:
  //   choose a random point
  //   choose an amplitude
  //   raise terrain in circular pattern to reach amplitude

  Real cavex = getWidth()/2;
  Real cavez = getLength()-20;

  int numDunes = slowMode? 200 : 2000;
  while ( numDunes-- ) {
    // Starting point
    int rdune, cdune;
    randomRowCol ( rows, cols, rdune, cdune );
    int idune = mapIndex(rdune,cdune);
    
    // Keep away from the cave
    GLfloat dx, dz; unmapPos ( rdune, cdune, dx, dz );
    if ( fabs(dx - cavex) + fabs(dz - cavez) < 150 )
      continue;

    // Amplitude
    Real amp = 1.0 + .15*(rand() % 10); // 1 - 2;
    Real finalHeight = map[idune] + amp;

    int radius = int(10 * amp);
    while ( map[idune] < finalHeight ) {
      Real dh = amp * .1* (rand()%10);
      for ( int r = -radius; r <= radius; ++r ) {
	int r2 = r*r;
	for ( int c = -radius; c <= radius; ++c ) {
	  int c2 = c*c;
	  Real f = 1.0;
	  if ( r2 || c2 ) f /= pow(r2 + c2, .3);
	  incr ( rdune + r, cdune + c, f*dh );
	}
      }
    }

  }
}



/**********************************************************************
 * GroundPatch
 **********************************************************************/


Texture* GroundPatch::wallTex = 0;
#include "rock1.c" // Wall texture image


GroundPatch::GroundPatch ( Ground* gr, int r1, int c1, int r2, int c2 )
  : PosObj(), DisplayBuilder(),
    ground(gr), didDivision(false),
    r1(r1), c1(c1), r2(r2), c2(c2), patch(),
    dlist(0)
{
  // Determine radius
  int dc = c2 - c1;
  int dr = r2 - r1;
  GLfloat w = dc * ground->dx;
  GLfloat h = dr * ground->dz;
  GLfloat rad = w > h? w/2.0 : h/2.0;

  setBoundRadius ( rad * 1.5 ); // Prevent false culling by extra boundary


  // Set position -- assume map pos (0,0) is at origin
  GLfloat midx = (c1 + c2) * ground->dx/2.0;
  GLfloat midz = (r1 + r2) * ground->dz/2.0;
  setPos ( Vec3 ( midx, 0, midz ) );

  patch[0] = patch[1] = patch[2] = patch[3] = 0;


  // Wall texture
  if ( !wallTex && !ground->slowMode ) {
    wallTex = new Texture;
    wallTex->buildMipmaps ( rock1_img.width,
			    rock1_img.height,
 			    rock1_img.pixel_data );
    wallTex->unbind();
  }
}


GroundPatch::~GroundPatch ()
{
  delete patch[0];
  delete patch[1];
  delete patch[2];
  delete patch[3];
  delete dlist;
}


void GroundPatch::divide ()
{
  int rm = (r1 + r2)/2;
  int cm = (c1 + c2)/2;

  patch[0] = new GroundPatch ( ground, r1, c1, rm, cm );
  patch[1] = new GroundPatch ( ground, r1, cm, rm, c2 );
  patch[2] = new GroundPatch ( ground, rm, c1, r2, cm );
  patch[3] = new GroundPatch ( ground, rm, cm, r2, c2 );
}


void GroundPatch::redraw ( int rr1, int rc1, int rr2, int rc2 )
{
  if ( r2 < rr1 || r1 > rr2 ||
       c2 < rc2 || c1 > rc2 )
    return; // No intersection

  if ( dlist ) {
//     cerr << "redraw patch " << this
// 	 << " r1=" << r1 << " c1= "<< c1
// 	 << " r2=" << r2 << " c2= "<< c2 << "\n";
    delete dlist;
    dlist = 0;
  }

  if ( patch[0] ) {
    patch[0]->redraw ( rr1, rc1, rr2, rc2 );
    patch[1]->redraw ( rr1, rc1, rr2, rc2 );
    patch[2]->redraw ( rr1, rc1, rr2, rc2 );
    patch[3]->redraw ( rr1, rc1, rr2, rc2 );
  }

}



void GroundPatch::setGndMaterial ( GLfloat y, GLfloat maxy, int r, int c )
{
  GLfloat fy = y / maxy;
  if ( fy < 0 ) fy = 0;
  if ( fy > 1 ) fy = 1;

  // Material properties for lighted rendering
  // Mix two colors together depending on height
  // Would be nice to make this textured, but maybe later

  // Tan
  static const Vec4 gndAmbient1 ( .824, .706, .549, 1.0 );

  // Darker tan
  static const Vec4 gndAmbient2 = gndAmbient1 * .3;

  // Black
  static const Vec4 gndAmbient3 ( .2, .1, .1, 1.0 );

  Vec4 mixAmbient;
  if ( ground->expMap[ground->mapIndex(r,c)] ) {
    // explosion
    mixAmbient = gndAmbient2*(fy) + gndAmbient3*(1.0-fy);
  } else {
    // normal
    mixAmbient = gndAmbient1*(fy) + gndAmbient2*(1.0-fy);
  }

//   if ( c == ground->cols / 2 ) mixAmbient = Vec4 ( 0, 1, 0, 1 );

  glMaterialfv ( GL_FRONT, GL_AMBIENT_AND_DIFFUSE, mixAmbient );
  glMaterialfv ( GL_FRONT, GL_SPECULAR, mixAmbient*.2 );
  glMaterialf ( GL_FRONT, GL_SHININESS, 0 );

  if ( wallTex ) wallTex->unbind();
}


void GroundPatch::setWallMaterial ( GLfloat y )
{
  GLfloat fy = y / wallHeight;
  if ( fy > 1.0 ) fy = 1.0;
//   cerr << "y = " << y << " --> fy = " << fy << "\n";

  // Material properties for lighted rendering
  // Mix two colors together depending on height
  // Would be nice to make this textured, but maybe later

//   // Wheat
//   static Vec4 gndAmbient1 ( .961, .871, .702, 1.0 );
  // Tan
  static const Vec4 gndAmbient1 ( .824, .706, .549, 1.0 );

  // Brown
  static const Vec4 gndAmbient2 ( .4, .165, .165, 1.0 );

  Vec4 mixAmbient = gndAmbient1*(1.0-fy) + gndAmbient2*(fy);

  glMaterialfv ( GL_FRONT, GL_AMBIENT, gndAmbient2 );
  glMaterialfv ( GL_FRONT, GL_DIFFUSE, mixAmbient );
  glMaterialfv ( GL_FRONT, GL_SPECULAR, mixAmbient*.6 );
  glMaterialf ( GL_FRONT, GL_SHININESS, 10 );

  if ( wallTex ) wallTex->bind();
}


static Real texRand () { return -.7 + 0.01 * (rand()%70); }

void GroundPatch::buildDisplayList ()
{
  glPushAttrib ( GL_EVAL_BIT | GL_ENABLE_BIT );
  glShadeModel ( ground->slowMode? GL_FLAT : GL_SMOOTH );
  glPolygonMode ( GL_FRONT_AND_BACK, GL_FILL );

  glEnable ( GL_AUTO_NORMAL );
  glBegin ( GL_QUADS );
  int dr = (r2 - r1)/8;
  int dc = (c2 - c1)/2;

  GLfloat basex, basez;
  ground->unmapPos ( r1, c1, basex, basez );

  for ( int r0=r1; r0 < r2; r0 += dr ) {
    int r = r0 + dr;
    GLfloat z0 = r0*ground->dz;
    GLfloat z1 = r*ground->dz;

    for ( int c0=c1; c0 < c2; c0 += dc ) {
      int c = c0 + dc;
      GLfloat x0 = c0*ground->dx;
      GLfloat x1 = c*ground->dx;

      GLfloat y0 = ground->map[ground->mapIndex(r0,c0)];
      GLfloat y1 = ground->map[ground->mapIndex(r,c0)];
      GLfloat y2 = ground->map[ground->mapIndex(r,c)];
      GLfloat y3 = ground->map[ground->mapIndex(r0,c)];
      
      Vec3 v0 ( x0, y0, z0 );
      Vec3 v1 ( x0, y1, z1 );
      Vec3 v2 ( x1, y2, z1 );
      Vec3 v3 ( x1, y3, z0 );

//       cerr << "r0=" << r0 << " c0=" << c0 << " v0=" << v0 << "\n";

      bool isWall = ( ground->isWall(r0,c0) &&
		      ground->isWall(r,c0) &&
		      ground->isWall(r,c) &&
		      ground->isWall(r0,c) );
//       if ( isWall ) continue;
      if ( isWall ) setWallMaterial ( y0 );
      else {
	glTexCoord2f ( 0,0 );
      }

      if ( isWall ) { glTexCoord2f ( 0+texRand(), 0+texRand() ); }
      else setGndMaterial ( y0, ground->maxy, r0, c0 );
      glVertex3fv ( v0 );

      if ( isWall ) { glTexCoord2f ( 0+texRand(), 1+texRand() ); }
      else setGndMaterial ( y1, ground->maxy, r, c0 );
      glVertex3fv ( v1 );

      if ( isWall ) { glTexCoord2f ( 1+texRand(), 1+texRand() ); }
      else setGndMaterial ( y2, ground->maxy, r, c );
      glVertex3fv ( v2 );

      if ( isWall ) { glTexCoord2f ( 1+texRand(), 0+texRand() ); }
      else setGndMaterial ( y3, ground->maxy, r0, c );
      glVertex3fv ( v3 );
    }
  }      
  glEnd();

  glPopAttrib();

  if ( wallTex ) wallTex->unbind();
}


void GroundPatch::render ( Viewport* vp )
{
//   cerr << "\nGround::render()\n";
  _render ( vp ); // Avoid virtual function call on recursion
//   cerr << "  done Ground::render()\n";
//   exit(0);
}


bool GroundPatch::_render ( Viewport* vp )
{
  if ( !vp->canSee ( this ) ) {
//     cerr << "Can't see patch " << this
// 	 << " at " << getPos() << " rad=" << getBoundRadius() << "\n";
    return false;
  }
  
//   Real dist = len( getPos() - vp->getPos() );
//   cerr << "Visible patch " << this
//        << " at " << getPos() << " rad=" << getBoundRadius()
//        << " dist=" << dist << "\n";

  // Draw at this level or recurse?
  // If close in, recurse to get finer level of detail
  bool drew = false;
  if ( !didDivision ) {
    // Divide down as necessary
    int dc = c2 - c1;
    int dr = r2 - r1;
    static const int divideThresh = 16;
    if ( dc >= divideThresh || dr >= divideThresh )
      divide();
    didDivision = true;
  }
  
  // Currently only draws leaf-level patches!
  // If iti s fast enough, keep it that way.
  if ( patch[0] ) {
    //     cerr << "  Recursing on patch " << this << "\n";
    if ( patch[0]->_render ( vp ) ) drew = true;
    if ( patch[1]->_render ( vp ) ) drew = true;
    if ( patch[2]->_render ( vp ) ) drew = true;
    if ( patch[3]->_render ( vp ) ) drew = true;
  }

  if ( !drew ) {
    if ( !dlist ) {
      dlist = buildNew( true );
    }
    
    dlist->call();
    drew = true;
  }

  return drew;
}


} // namespace CS148
