// Mover.C

#include "Mover.H"
#include "Ground.H"


namespace CS148 {


/**********************************************************************
 * Mover
 **********************************************************************/


Mover::Mover ()
  : PosObj(),
    velo(), rotation(0), maxSpeed(0), minSpeed(0),
    ground(0), targetHeight(0), descentRate(1.0),
    turns(),
    bounded(false), x1(0), z1(0), x2(0), z2(0)
{


}


Mover::~Mover ()
{

}


void Mover::move ( Real interval )
{
  adjustHeight ( interval );
  applyVelo ( interval );
  applyTurns ( interval );
  checkBoundary ();
}


void Mover::setVelo ( const Vec3& v )
{
  velo = v;

  // Get rotation about y in degrees, based on velocity
  // (assumes always facing in direction of travel)
  Vec3 d = norm(velo);
  rotation = rad2deg( arctan( d[0], d[2] ) );

//   cerr << "setVelo: velo=" << getVelo()
//        << ", rot=" << rotation << "\n";

}


void Mover::setElevation ( Ground* g, Real el, Real dr )
{
  ground = g;
  targetHeight = el;
  descentRate = dr;

  Vec3 pos = getPos();
  pos[1] = ground->getHeight( pos[0], pos[2] ) + targetHeight;
  setPos ( pos );
}


void Mover::setBoundary ( Real x1, Real z1, Real x2, Real z2 )
{
  this->x1 = x1;
  this->z1 = z1;
  this->x2 = x2;
  this->z2 = z2;
  bounded = true;
}
  

Vec3 Mover::getPosAfter ( Real interval )
{
  // predict future position, just based on current
  // velocity (not taking turns into consideration)
  return getPos() + getVelo() * interval;
}


void Mover::addTurn ( Turn* t )
{
  // Limit number of simultaneous turns
  if ( turns.size() == maxTurns ) {
    delete turns[0];
    turns.erase ( turns.begin() );
  }

  turns.push_back ( t );
}


void Mover::adjustHeight ( Real interval )
{
//   cerr << "targetHeight = " << targetHeight
//        << " height = " << getPosAfter(1.0)[1] << "\n";

  if ( targetHeight == 0 ) return;

  // Look at height ahead a bit
  Real future = .3;
  Vec3 fpos = getPosAfter ( future );
  Real gheight = ground->getHeight ( fpos[0], fpos[2] );

  // Adjust height gradually, based on the change in height
  // relative to where we are now
  Real ftargetHeight = targetHeight + gheight;
  Real missHeight = ftargetHeight - fpos[1]; // miss positive when too low
  Real dy = missHeight * 5* interval;
  if ( dy < 0 ) dy *= descentRate;
//   cerr << "fheight = " << ftargetHeight << "  fy = " << fpos[1]
//        << "  missHeight = " << missHeight << " dy = " << dy << "\n";

  Vec3 pos = getPos();
  pos[1] += dy;
  if ( pos[1] < targetHeight/2.0 ) pos[1] = targetHeight/2.0;
  setPos ( pos );
}


void Mover::applyTurns ( Real interval )
{
  if ( turns.size() == 0 ) return;

  Vec3 newv = getVelo();
  bool completed = false;

  // Change velocity
  for ( vector<Turn*>::iterator iter = turns.begin();
	iter != turns.end(); ++iter ) {
    Turn* turn = *iter;
    newv = turn->apply( newv, interval );

    if ( turn->completed() ) {
      delete turn;
      *iter = 0;
      completed = true;
    }
  }

  setVelo ( newv );


  // Remove completed turns from the list
  if ( completed ) {
    do {
      completed = false;
      for ( vector<Turn*>::iterator iter = turns.begin();
	    iter != turns.end(); ++iter ) {
	if ( !*iter ) {
	  turns.erase ( iter );
	  completed = true;
	  break;
	}
      }
    } while ( completed );
  }

}


void Mover::applyVelo ( Real interval )
{
  setPos ( getPos() + getVelo() * interval );
}


void Mover::turn ( Real deg, Real rate )
{
  addTurn ( new Turn ( deg, deg>0? rate : -rate ) );
}


void Mover::accelerate ( Real dspeed )
{
  Vec3 newv = getVelo();
  Vec3 dir = norm(newv);
  Real speed = len(newv);
  if ( speed == 0 ) dir = Vec3( 1, 0, 0 );

  speed += dspeed;
  if ( speed > maxSpeed ) speed = maxSpeed;
  else if ( speed < minSpeed ) speed = minSpeed;

  setVelo ( dir * speed );
}


void Mover::checkBoundary ()
{
  // Check for proximity to boundary, and automatically
  // turn away from it
  if ( !bounded ) return;

  Vec3 fpos = getPosAfter ( .5 );

  bool doTurn = false;
  Vec3 wall, nrm;

  if ( fpos[0] < x1 ) {
    // Past east boundary
    wall = Vec3 ( 0, 0, 1 );
    nrm = Vec3 ( 1, 0, 0 );
    doTurn = true;
  }
  else if ( fpos[0] > x2 ) {
    // Past west boundary
    wall = Vec3 ( 0, 0, -1 );
    nrm = Vec3 ( -1, 0, 0 );
    doTurn = true;
  }
  else if ( fpos[2] < z1 ) {
    // Past south boundary
    wall = Vec3 ( -1, 0, 0 );
    nrm = Vec3 ( 0, 0, 1 );
    doTurn = true;
  }
  else if ( fpos[2] > z2 ) {
    // Past north boundary
    wall = Vec3 ( 1, 0, 0 );
    nrm = Vec3 ( 0, 0, -1 );
    doTurn = true;
  }

  if ( doTurn && dot( getVelo(), nrm ) < 0 ) {
    Real d = dot ( norm(getVelo()), wall );
    Real turnDeg = 20;
//     cerr << "turn from wall=" << wall
// 	 << " turnDeg=" << turnDeg << " d=" << d << "\n";
    if ( d < 0 ) { d *= -1; turnDeg *= -1; }
    Real rate = fabs(2*turnDeg);
    turnDeg *= 1.2 - d;
//     cerr << ": deg=" << turnDeg << " rate=" << rate << "\n";
    turn ( turnDeg, rate );
  }

}


void Mover::warp ( Real x, Real z )
{
  // Debugging
  if ( ground ) {
    setPos ( Vec3( x,
		   ground->getHeight ( x, z ) + targetHeight,
		   z ) );
  }

}



} // namespace CS148

