#!/usr/bin/perl

use strict;
require './VectorReal.pm'; # Math::VectorReal


my $matNum = 0;
my $obj = undef;
my @objs = ();

while ( <> ) {

  if ( /^MATERIAL \"(.+)\" rgb (.+)  amb (.+)  emis (.+)  spec (.+)  shi (.+)  trans (.+)/ ) {
    my ($matname, $rgb, $amb, $emis, $spec, $shi, $trans) =
      ($1, $2, $3, $4, $5, $6, $7);
    my $alpha = 1.0 - $trans;

    printf ( "static void setMaterial_%d ()\n".
	     "{\n".
	     "  static const GLfloat diff[4] = {%.3f,%.3f,%.3f,%.3f};\n".
	     "  static const GLfloat amb[4] = { %.3f, %.3f, %.3f, %.3f };\n".
	     "  static const GLfloat emis[4] = { %.3f, %.3f, %.3f, %.3f };\n".
	     "  static const GLfloat spec[4] = { %.3f, %.3f, %.3f, %.3f };\n".
	     "  static const GLfloat shi = %.3f;\n".
	     "\n".
	     "  glMaterialfv ( GL_FRONT_AND_BACK, GL_DIFFUSE, diff );\n".
	     "  glMaterialfv ( GL_FRONT_AND_BACK, GL_AMBIENT, amb );\n".
	     "  glMaterialfv ( GL_FRONT_AND_BACK, GL_EMISSION, emis );\n".
	     "  glMaterialfv ( GL_FRONT_AND_BACK, GL_SPECULAR, spec );\n".
	     "  glMaterialf ( GL_FRONT_AND_BACK, GL_SHININESS, shi );\n".
	     "}\n\n",

	     $matNum++,
	     split(' ', $rgb), $alpha,
	     split(' ', $amb), $alpha,
	     split(' ', $emis), $alpha,
	     split(' ', $spec), $alpha,
	     $shi );
    next;
  }

  if ( /^OBJECT poly/ ) {
    $obj = {};
    push @objs, $obj;
    next;
  }

  next unless $obj;
  
  if ( /^name \"(.+)\"/ ) {
    $obj->{name} = $1;
    next;
  }

  if ( /^loc (\S+) (\S+) (\S+)/ ) {
    $obj->{loc} = new Math::VectorReal ( $1, $2, $3 );
    next;
  }

  if ( /^numvert (\d+)/ ) {
    my $numvert = $1;
    $obj->{verts} = [];
    while ( $numvert-- ) {
      $_ = <>;
      push @{$obj->{verts}}, new Math::VectorReal ( split ( ' ', $_ ) );
    }
    next;
  }

  if ( /^numsurf (\d+)/ ) {
    my $numsurf = $1;
    $obj->{surfs} = [];
    next;
  }

  if ( /^SURF (\S+)/ ) {
    $obj->{surf} = { flag => $1 };
    push @{$obj->{surfs}}, $obj->{surf};
    next;
  }

  if ( /^mat (\d+)/ ) {
    $obj->{surf}{mat} = $1;
    next;
  }

  if ( /^refs (\d+)/ ) {
    my $numvert = $1;
    $obj->{surf}{verts} = [];
    while ( $numvert-- ) {
      $_ = <>;
      my ($i,$u,$v) = split(' ', $_);
      push @{$obj->{surf}{verts}}, [ $i, $u, $v ];
    }
    next;
  }

}



# Do surface normals - Newell Method
foreach my $obj ( @objs ) {
  my $verts = $obj->{verts};
  my $vertsurfs = $obj->{vertsurfs} = [];
  foreach my $surf ( @{$obj->{surfs}} ) {
    my $sverts = $surf->{verts};
    my @vecs = ();
    foreach my $svert ( @$sverts ) {
      my $i = $svert->[0];
      push @vecs, $i;
      $vertsurfs->[$i] = [] unless $vertsurfs->[$i];
      push @{$vertsurfs->[$i]}, $surf;
    }

    my ($nx, $ny, $nz) = (0,0,0);
    for ( my $i = 0; $i < scalar(@vecs); ++$i ) {
      my $j = ($i + 1) % scalar(@vecs);
      my $vi = $obj->{verts}[$i];
      my $vj = $obj->{verts}[$j];
      $nx += ( ($vi->y() - $vj->y()) * 
	       ($vi->z() + $vj->z()) );
      $ny += ( ($vi->z() - $vj->z()) * 
	       ($vi->x() + $vj->x()) );
      $nz += ( ($vi->x() - $vj->x()) * 
	       ($vi->y() + $vj->y()) );
    }
    $surf->{normal} = new Math::VectorReal ( $nx, $ny, $nz ) -> norm;
  }

  # Now vertex normals by averaging surface normals
  $obj->{normals} = [];
  for ( my $i = 0; $i < scalar(@{$obj->{verts}}); ++ $i ) {
    my $norm = new Math::VectorReal (0,0,0);
    foreach my $surf ( @{$vertsurfs->[$i]} ) {
      $norm += $surf->{normal};
    }
    $obj->{normals}[$i] = $norm->norm();
  }
}


# Print each obj
# Don't use vertex arrays, since they don't work with DispayLists.

$Math::VectorReal::FORMAT = "Vec3( %.5f, %.5f, %.5f )";

foreach my $obj ( @objs ) {

  my $name = $obj->{name};
  my $loc = $obj->{loc};
  my $verts = $obj->{verts};
  my $norms = $obj->{normals};

  printf ( "\n\nclass $name : public DisplayBuilder\n".
	   "{\n".
	   "public:\n".
	   "  $name () : DisplayBuilder() {}\n".
	   "  ~$name () {}\n".
	   "  \n".
	   "  void buildDisplayList () {\n".
	   "    glPushMatrix();\n".
	   "    glPolygonMode ( GL_FRONT_AND_BACK, GL_FILL );\n".
	   "    glEnable ( GL_BLEND );\n".
	   "    glBlendFunc ( GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA );\n".
	   "  \n" );

  printf ( "    glTranslatef ( %.5f, %.5f, %.5f );\n",
	   $loc->x(), $loc->y(), $loc->z() );

  # Print all the verts, normals, and tex coords together

  my $curMat = -1;
  my $curMode = undef;
  foreach my $surf ( @{$obj->{surfs}} ) {
    my $setMat = undef;
    if ( $surf->{mat} != $curMat ) {
      $setMat = $surf->{mat};
    }

    my $sverts = $surf->{verts};
    my $mode = 'GL_TRIANGLES';
    if ( @$sverts > 3 ) {
      $mode = 'GL_POLYGON';
    }
    my $changeMode = ( $mode ne $curMode ) || $setMat;
    if ( $changeMode ) {
      if ( defined ( $curMode ) ) { printf ( "    glEnd();\n\n" ); }
    }


    # Material
    if ( defined $setMat ) {
      printf ( "    setMaterial_%d();\n", $setMat );;
      $curMat = $setMat;
    }

    # The vertices
    if ( $changeMode ) {
      printf ( "    glBegin ( $mode );\n" );
      $curMode = $mode
    }

    for ( my $i = 0; $i < scalar(@$sverts); ++$i ) {
      my ($vi, $texu, $texv) = @{$sverts->[$i]};
      my $vert = $verts->[$vi];
      my $norm = $norms->[$vi];

      printf ( "      glTexCoord2f ( %.5f, %.5f );\n", $texu, $texv );
      printf ( "      glNormal3fv ( %s );\n", $norm );
      printf ( "      glVertex3fv ( %s );\n", $vert );
    }

    if ( $mode eq 'GL_POLYGON' ) {
      printf ( "    glEnd();\n" );
      $curMode = undef;
    }
    else { printf ( "    \n" ); }
  }

  if ( defined $curMode ) {
    printf ( "    glEnd();\n" );
  }

  printf ( "    glPopMatrix();\n" );
  printf ( "    glDisable ( GL_BLEND );\n" );
  printf ( "\n".
	   "  }\n".
	   "};\n\n" );
}



